/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::Navigator
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "Navigator.h"

#include <algorithm>
#include <cmath>
#include <iterator>
#include <memory>
#include <utility>

#include <Eigen/Geometry>

#include <Ice/Current.h>

#include <SimoxUtility/math/convert/rpy_to_mat3f.h>
#include <VirtualRobot/MathTools.h>
#include <VirtualRobot/Robot.h>

#include <ArmarXCore/core/exceptions/LocalException.h>
#include <ArmarXCore/core/logging/LogSender.h>
#include <ArmarXCore/core/services/tasks/PeriodicTask.h>
#include <ArmarXCore/core/services/tasks/TaskUtil.h>
#include <ArmarXCore/util/CPPUtility/trace.h>

#include "ArmarXGui/libraries/RemoteGui/Client/Widgets.h"

#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>

#include <armarx/navigation/client/NavigationStackConfig.h>
#include <armarx/navigation/components/Navigator/RemoteGui.h>
#include <armarx/navigation/core/types.h>
#include <armarx/navigation/factories/NavigationStackFactory.h>
#include <armarx/navigation/server/Navigator.h>
#include <armarx/navigation/server/execution/PlatformUnitExecutor.h>
#include <armarx/navigation/server/introspection/MemoryIntrospector.h>
#include <armarx/navigation/util/util.h>

std::vector<armarx::navigation::core::Pose>
convert(const std::vector<Eigen::Matrix4f>& wps)
{
    using namespace armarx::navigation;
    std::vector<core::Pose> p;
    p.reserve(wps.size());
    std::transform(
        wps.begin(), wps.end(), std::back_inserter(p), [](const auto& p) { return core::Pose(p); });
    return p;
}

armarx::navigation::components::Navigator::Navigator() :
    parameterizationReader(memoryNameSystem),
    parameterizationWriter(memoryNameSystem),
    eventsWriter(memoryNameSystem),
    resultsWriter(memoryNameSystem),
    parameterizationService(&parameterizationReader, &parameterizationWriter),
    publisher(&resultsWriter, &eventsWriter)
{
    scene.timeServer = &timeServer;
}


armarx::navigation::components::Navigator::~Navigator() = default;

void
armarx::navigation::components::Navigator::onInitComponent()
{
}

void
armarx::navigation::components::Navigator::onConnectComponent()
{
    static bool initialized{false};

    // redirect to onReconnect to avoid any setup issues
    if (initialized)
    {
        ARMARX_INFO << "Reconnecting ...";
        onReconnectComponent();
        return;
    }

    // TODO check if reconnecting

    scene.robot = getRobot();
    scene.staticScene = staticScene();

    executor = server::PlatformUnitExecutor(platformUnit);

    introspector = server::ArvizIntrospector(getArvizClient(), scene.robot);

    // TODO dynamic scene
    // TODO memory
    // TODO param (10)
    resultsWriter.connect();
    eventsWriter.connect();
    // parameterizationReader.connect();
    parameterizationWriter.connect();


    robotStateUpdateTask = new PeriodicTask<Navigator>(
        this, &Navigator::updateRobot, 10, false, "RobotStateUpdateTask");
    robotStateUpdateTask->start();

    navRemoteGui = std::make_unique<NavigatorRemoteGui>(remoteGui, *this);
    navRemoteGui->enable();


    initialized = true;
}

void
armarx::navigation::components::Navigator::onReconnectComponent()
{
    robotStateUpdateTask->start();

    // TODO not in all cases meaningful
    //resume();
    resultsWriter.connect();
    eventsWriter.connect();
    // parameterizationReader.connect();
    parameterizationWriter.connect();

    navRemoteGui->enable();
}

void
armarx::navigation::components::Navigator::onDisconnectComponent()
{
    robotStateUpdateTask->stop();

    stopAll();

    navRemoteGui->disable();
}

void
armarx::navigation::components::Navigator::onExitComponent()
{
}

void
armarx::navigation::components::Navigator::updateContext()
{
    scene.robot = getRobot();
    scene.staticScene = staticScene();
    // TODO dynamic scene
}

std::string
armarx::navigation::components::Navigator::getDefaultName() const
{
    return "Navigator";
}

void
armarx::navigation::components::Navigator::createConfig(const aron::data::AronDictPtr& stackConfig,
                                                        const std::string& callerId,
                                                        const Ice::Current&)
{
    // TODO: Not thread-safe.
    ARMARX_TRACE;
    ARMARX_INFO << "Creating config for caller '" << callerId << "'";

    parameterizationService.store(stackConfig, callerId, timeServer.now());

    server::NavigationStack stack = fac::NavigationStackFactory::create(stackConfig, scene);

    memoryIntrospectors.emplace_back(
        std::make_unique<server::MemoryIntrospector>(resultsWriter, callerId));

    navigators.emplace(
        std::piecewise_construct,
        std::forward_as_tuple(callerId),
        std::forward_as_tuple(
            server::Navigator::Config{.stack = std::move(stack),
                                      .scene = &scene,
                                      .general = server::Navigator::Config::General{}},
            server::Navigator::InjectedServices{.executor = &executor.value(),
                                                .publisher = &publisher,
                                                .introspector = &(introspector.value())}));
}

void
armarx::navigation::components::Navigator::moveTo(const std::vector<Eigen::Matrix4f>& waypoints,
                                                  const std::string& navigationMode,
                                                  const std::string& callerId,
                                                  const Ice::Current&)
{

    ARMARX_INFO << "moveTo() requested by caller '" << callerId << "'";

    try
    {
        navigators.at(callerId).moveTo(convert(waypoints),
                                       core::NavigationFrameNames.from_name(navigationMode));
    }
    catch (...)
    {
        // TODO: Error handling.
        ARMARX_WARNING << "Failed to execute moveTo()."
                       << "Movement will be paused.";
        stopAll(); // TODO pause movement must not throw
        throw;
    }
}

void
armarx::navigation::components::Navigator::moveTowards(const Eigen::Vector3f& direction,
                                                       const std::string& navigationMode,
                                                       const std::string& callerId,
                                                       const Ice::Current&)
{
    // TODO: Error handling.

    ARMARX_INFO << "MoveTowards requested by caller '" << callerId << "'";

    navigators.at(callerId).moveTowards(direction,
                                        core::NavigationFrameNames.from_name(navigationMode));
}

void
armarx::navigation::components::Navigator::pause(const std::string& configId, const Ice::Current&)
{
    navigators.at(configId).pause();
}

void
armarx::navigation::components::Navigator::resume(const std::string& configId, const Ice::Current&)
{
    navigators.at(configId).resume();
}

void
armarx::navigation::components::Navigator::stop(const std::string& configId, const Ice::Current&)
{
    navigators.at(configId).stop();
}

void
armarx::navigation::components::Navigator::stopAll(const Ice::Current&)
{
    for (auto& [_, navigator] : navigators)
    {
        navigator.stop();
    }
}

bool
armarx::navigation::components::Navigator::isPaused(const std::string& configId,
                                                    const Ice::Current&)
{
    return navigators.at(configId).isPaused();
}

bool
armarx::navigation::components::Navigator::isStopped(const std::string& configId,
                                                     const Ice::Current&)
{
    return navigators.at(configId).isStopped();
}

armarx::PropertyDefinitionsPtr
armarx::navigation::components::Navigator::createPropertyDefinitions()
{
    ARMARX_TRACE;

    PropertyDefinitionsPtr def = new ComponentPropertyDefinitions(getConfigIdentifier());

    // Publish to a topic (passing the TopicListenerPrx).
    // def->topic(myTopicListener);

    // Subscribe to a topic (passing the topic name).
    // def->topic<PlatformUnitListener>("MyTopic");

    // Use (and depend on) another component (passing the ComponentInterfacePrx).
    def->component(platformUnit);

    def->component(remoteGui, "RemoteGuiProvider");
    // Add a required property.
    // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

    // Add an optionalproperty.
    // def->optional(properties.numBoxes, "p.box.Number", "Number of boxes to draw in ArViz.");

    resultsWriter.registerPropertyDefinitions(def);
    eventsWriter.registerPropertyDefinitions(def);
    // parameterizationReader.registerPropertyDefinitions(def);
    parameterizationWriter.registerPropertyDefinitions(def);

    return def;
}

armarx::navigation::core::StaticScene
armarx::navigation::components::Navigator::staticScene()
{
    ARMARX_TRACE;

    const objpose::ObjectPoseSeq objectPoses = ObjectPoseClientPluginUser::getObjectPoses();
    core::StaticScene scene{.objects = util::asSceneObjects(objectPoses)};

    ARMARX_INFO << "The scene consists of " << scene.objects->getSize() << " objects";

    return scene;
}

VirtualRobot::RobotPtr
armarx::navigation::components::Navigator::getRobot()
{
    ARMARX_TRACE;

    auto robot = RemoteRobot::createLocalCloneFromFile(
        getRobotStateComponent(), VirtualRobot::RobotIO::RobotDescription::eFull);
    // auto robot = RemoteRobot::createLocalClone(getRobotStateComponent());
    ARMARX_CHECK_NOT_NULL(robot);

    // ARMARX_INFO << "Resetting robot";
    // robot = VirtualRobot::RobotIO::loadRobot(robot->getFilename());

    RemoteRobot::synchronizeLocalClone(robot, getRobotStateComponent());

    return robot;
}

void
armarx::navigation::components::Navigator::updateRobot()
{
    synchronizeLocalClone(scene.robot);
}

armarx::navigation::server::Navigator*
armarx::navigation::components::Navigator::activeNavigator()
{
    // We define the active navigator to be the one whose movement is enabled.
    const auto isActive = [](auto& nameNavPair) -> bool
    {
        server::Navigator& navigator = nameNavPair.second;
        return not navigator.isPaused();
    };

    auto it = std::find_if(navigators.begin(), navigators.end(), isActive);

    // no navigator active?
    if (it == navigators.end())
    {
        return nullptr;
    }

    return &it->second;
}
